import matplotlib.pyplot as plt
import numpy as np


def rotate_2D(X, Y, theta):
    """
    Simple 2D rotation
    """

    X1 = X*np.cos(theta) - Y*np.sin(theta)
    X2 = X*np.sin(theta) + Y*np.cos(theta)

    return [X1, X2]

def get_ellipse2D(var_maj, var_min, theta, n):
    """
    Get coordinates of an n point ellipse on an x-y plane with major axis at anb angle of theta. var_maj 
    is the length of the major axis and var_min is the length of the minor axis. 
    
    """
    t = np.linspace(0, 2*np.pi, n);
    x = var_maj*np.cos(t)*np.cos(theta) - var_min*np.sin(t)*np.sin(theta);
    y = var_maj*np.cos(t)*np.sin(theta) + var_min*np.sin(t)*np.cos(theta);

    return x, y

def PCA_2D(X, Y, ellipse_stdevs=2, plotopt=False):
    """
    PCA on a plane (2D)
    
    Originally from matlab code of Jeff Book's masterclass, modified by Tamara Schlosser, and then by Andrew Zulberti to 
    be more similar to the Emery and Thompson notation. 
    
    TO DO: generalise to 3D
    
    Inputs:
        X - timeseries 1 - on arbitrary axis x
        Y - timeseries 2 - on axis perpendicular to x
        
    Outputs:
        theta - radian angle of PC measured counterclockwise from axis x
        x     - x coordinates of ellipse
        y     - y coordinates of ellipse
    """
    
    # %% Now attempting the notation in Emery and Thompson

    X = X[~np.isnan(X)];
    Y = Y[~np.isnan(Y)];

    mean_x = np.mean(X);
    mean_y = np.mean(Y);

    u_1 = X - mean_x;
    u_2 = Y - mean_y;

    c = np.cov(u_1, u_2);
    cov_12 = c[0, 1];
    var_1  = c[0, 0];
    var_2  = c[1, 1];

    # % det|C - lambda*I| = 0
    # % using this in the quadratic equation gives
    lambda_1 = (1/2)*((var_1+var_2)+np.sqrt((var_1-var_2)**2 + 4*cov_12**2));
    lambda_2 = (1/2)*((var_1+var_2)-np.sqrt((var_1-var_2)**2 + 4*cov_12**2));

    theta_p = (1/2)*np.arctan2(2*cov_12, (var_1 - var_2));

    sqrt_lambda_1 = np.sqrt(lambda_1);
    sqrt_lambda_2 = np.sqrt(lambda_2);

    x, y = get_ellipse2D(ellipse_stdevs*sqrt_lambda_1, ellipse_stdevs*sqrt_lambda_2, theta_p, 100)
#     x, y = get_ellipse2D(sqrt_lambda_1, sqrt_lambda_2, theta_p, 100)
    
    ellipse = [x, y]
    sqrt_lambda = np.array([sqrt_lambda_1, sqrt_lambda_2])
    X12 = rotate_2D(X, Y, -theta_p)

    if plotopt:
        
        n = len(X)

        print('Ellipse axis: {} deg'.format(theta_p*180/np.pi))
        
        plt.figure()
        plt.plot(X, Y, '.')
        plt.plot(x, y, 'r-')
        plt.title('n points: {} | Ellipse axis: {:0.2f} deg'.format(n, theta_p*180/np.pi))
    
    return theta_p, ellipse, sqrt_lambda, X12

def PCA_2D_eig(X, Y, ellipse_stdevs=2, plotopt=False):
    """
    Alternate method for 2D PCA using np.eig.

    This is more general and can be extended to higher dimensions than the algebraic method above.

    """

    sigma = np.cov(X, Y)

    w, v = np.linalg.eig(sigma)
    sqrt_lambda = np.sqrt(w)

    v1 = v[0, :] # First eigenvector
    theta_e = np.arctan2(v1[1], v1[0])
    theta_e = - theta_e # Just a convention thing

    x, y = get_ellipse2D(ellipse_stdevs*sqrt_lambda[0], ellipse_stdevs*sqrt_lambda[1], theta_e, 100)
    
    ellipse = [x, y]
    X12 = rotate_2D(X, Y, -theta_e)

    return theta_e, ellipse, sqrt_lambda, X12




import pandas as pd
import matplotlib.pyplot as plt 
import shapefile


def xyzbathy_2_contourshp(infiles, outfile, levels, field_name='DEPTH', ):
    """
    Function to convert xyz data into shapefiles of contours. Nominally it is for bathymetry, 
    a different field_name can be used for other scalars. 

    Inputs:
        infiles - xyx csv [or iterable of] with headers X, Y, Z [in capitals]
        outfile - name of the shapefile to be produced
        levels - the levels the contours are to be specified at.
        field_name - the field name that the scalar field (Z, column) will be written 
                to in the shp file. 
    """

    if type(infiles) == str:
        infiles = [infiles]
    
    CSs = []
    
    for infile in infiles:
        
        print('Contouring {}'.format(infile))
        
        df = pd.read_csv(infile)

        CS = plt.tricontour(df.X, df.Y, df.Z, 15, levels=levels, linewidths=0.5, colors='k')
        CSs += [CS]
        
    print('Done contouring.')
    plt.show()

    if False:
        for CS in CSs:
            for ii, segs in enumerate(CS.allsegs):

                print('{} m: {}'.format(levels[ii], len(segs)))

                for seg in segs:

                    plt.plot(seg[:, 0], seg[:, 1], 'k:', linewidth=0.5)

    with shapefile.Writer(outfile, shapeType=3) as w:
    
        w.field(field_name, 'N') # N = numbers
        
        for CS in CSs:
            for ii, segs in enumerate(CS.allsegs):
                print('{} m: {}'.format(levels[ii], len(segs)))

                for seg in segs:
                    plt.plot(seg[:, 0], seg[:, 1], 'k:', linewidth=0.5) 
                    w.line([seg])
                    w.record(DEPTH=levels[ii])
                    
    return CSs


import pandas as pd, xarray as xr
import numpy as np, matplotlib.pyplot as plt, datetime
import scipy.signal

def dirmath_to_dirgeo(dirmath, degrees=False):
    """
    Mathematical direction to geographical. Direction is in radians. 
    Direction is radians unless set degrees=True
    """
    
    return dirgeo_to_dirmath(dirmath, degrees=degrees)

def dirgeo_to_dirmath(dirgeo, degrees=False):
    """
    Mathematical direction to geographical. 
    Direction is radians unless set degrees=True
    """
    
    if degrees: 
        dirmath = 90 - dirgeo
    else: 
        dirmath = np.pi/2 - dirgeo
        
    return dirmath

def flip_dir(dire, degrees=False):
    """
    Flip direction 180 degrees/pi radians.
    Direction is radians unless set degrees=True
    """
    
    if degrees: 
        dire = 180 + dire
        dire = np.mod(dire, 360)
    else: 
        dire = np.pi + dire
        dire = np.mod(dire, 2*np.pi)
        
    return dire
    
def sd_to_uv(spd, dirgeo, convention, degrees=False):
    """
    Convert speed and direction to U and V. 
    Direction is clockwise from North. 
    Direction is radians unless set degrees=True
    """
        
    if not convention.lower() in ['meteo', 'ocean']:
        raise(Exception("Convention must be 'meteo' or 'ocean'"))
        
    if convention.lower() == 'meteo':
        dirgeo = flip_dir(dirgeo, degrees=degrees)
    
    dirmath = dirgeo_to_dirmath(dirgeo, degrees=degrees)
    
    if degrees:
        dirmath = dirmath*np.pi/180
        
    u = spd * np.cos(dirmath)
    v = spd * np.sin(dirmath)
    
    return u, v

def uv_to_sd(u, v, convention, degrees=False):
    """
    Convert U and V to speed and direction. 
    Direction is clockwise from North. 
    Direction is radians unless set degrees=True
    """
        
    if not convention.lower() in ['meteo', 'ocean']:
        raise(Exception("Convention must be 'meteo' or 'ocean'"))
        
    dirmath = np.arctan2(v, u)
    dirmath = np.mod(dirmath, np.pi*2)
    if degrees:
        dirmath = dirmath*180/np.pi
        
    spd = np.sqrt(u**2 + v**2)
    
    dirgeo = dirmath_to_dirgeo(dirmath, degrees=degrees)
    if convention.lower() == 'meteo':
        dirgeo = flip_dir(dirgeo, degrees=degrees)
        
    return spd, dirgeo



"""
Rapid implementation timeseries analysis. 

Collection of utilities for timeseries analysis with a focus on rapid implementation, not rapid runtimes. Code interfaces with numpy, scipy, xarray, matplotlib  

"""

import numpy as np, scipy.signal as signal, xarray as xr
import matplotlib.pyplot as plt

import afloat.time as ztime
import afloat.clean as zclean

import importlib, sys

############################################################
### Wanted to keep sfoda as a dependency here, but as I've 
### made a lot of changes it might be easier for now to use it 
### port it over here for now. Discuss with Matt. 
############################################################

# if (spec := importlib.util.find_spec('sfoda')) is not None:
#     module = importlib.util.module_from_spec(spec)
#     sys.modules['sfoda'] = module
#     spec.loader.exec_module(module)
# else:
#     raise(Exception(f"This function requires sfoda"))

# from sfoda.utils import harmonic_analysis as sha
from afloat.sfodautils import harmonic_analysis as sha

@xr.register_dataset_accessor("floatds")
class DataSet():
    """
    Accessor for xarray DataSet.

    Accessor for xarray DataSet which provides functionality for rapid implementation of timeseries analysis.

    Parameters
        ds : xarray.DataSet
            The DataSet to be accessed

    Returns
        ds : timeseries.DataSet
            The xarray accessor object

    Examples
        XXXXXXXX

    See Also
        xarray.DataSet
    """

    def __init__(self, ds): 
    
        self._obj = ds

    @property
    def _ds(self):
        return self._obj

    def __repr__(self):

        return self._ds.__repr__()

    def __getitem__(self, item):
        
        return DataArray(self._ds[item])

@xr.register_dataarray_accessor("floatda")
class DataArray():
    """
    Accessor for xarray DataArray.

    Accessor for xarray DataArray which provides functionality for rapid implementation of timeseries analysis.

    Parameters
        da : xarray.DataArray
            The DataArray to be accessed

    Returns
        da : timeseries.DataArray
            The xarray accessor object

    Examples
        XXXXXXXX

    See Also
        xarray.DataArray
    """

    def __init__(self, da):

        if 'units' in da.attrs:
            self.units = da.attrs['units']
        else:
            self.units = '?'
        
        dims = da.dims
        if not dims[0].lower() == 'time':
            raise(Exception("First dimension must be time"))

        self._obj = da

    @property
    def _da(self):

        return self._obj

    @property
    def dims(self):

        return self._obj.dims

    @property
    def other_dims(self):

        return [dim for dim in self._obj.dims if (not dim.lower()=='time')]

    @property
    def coords(self):
        return self._obj.coords

    def __repr__(self):

        return self._da.__repr__()

    @property
    def _timeseries(self):

        return TimeSeries(self._da.time.values, self._da.values, units=self.units, other_dims=self.other_dims)

    def plot(self, **kwargs):
        
        set_xlim = kwargs.pop("set_xlim", False)
        
        self._da.plot(**kwargs)
        
        ax = kwargs.pop("ax", plt.gca())

        if set_xlim:
            ax.set_xlim([min(self._da.time.values), max(self._da.time.values)])

    def interp(self, **kwargs):

        return self._timeseries.interp(**kwargs)

    def despike_gn02(self, **kwargs):

        return self._timeseries.despike_gn02(**kwargs)

    def butter(self, **kwargs):

        return self._timeseries.butter(**kwargs)

    def psd(self, **kwargs):

        return self._timeseries.psd(**kwargs)

    def psd_welch(self, **kwargs):

        return self._timeseries.psd_welch(**kwargs)

    def tidal_harmonic(self, Fin=None, **kwargs):

        return self._timeseries.tidal_harmonic(Fin=Fin, **kwargs)

    def rolling_tidal_harmonic(self, Fin=None, **kwargs):

        return self._timeseries.rolling_tidal_harmonic(Fin=Fin, **kwargs)

class TimeSeries():
    """
    An object to hold timeseries data and wrap the quick timeseries methods within this module. 
    """

    def __init__(self, time, data, other_dims=[], **kwargs):
        """

        """
        ## Start with no validation. Build organically. 
        
        quick_validate(time, data)

        units = kwargs.pop("units", '?')

        s = data.shape
        n_other_dims = len(s) - 1

        assert n_other_dims==len(other_dims), "If data is multidimensional, must specify names of other dimensions"
        assert len(time) == s[0], "Length of time must equal first dimension of data" 

        self.time = time
        self.data = data

        self.units = units

    def plot(self, ax=None, **kwargs):

        set_xlim = kwargs.pop("set_xlim", False)

        if ax is None:
            ax = plt.gca()

        ax.plot(self.time, self.data, **kwargs)
        ax.grid('on')

        if set_xlim:
            ax.set_xlim([min(self.time.values), max(self.time.values)])

    def interp(self, **kwargs):

        return quick_interp(self.time, self.data, **kwargs)

    def despike_gn02(self, **kwargs):

        return quick_despike_gn02(self.time, self.data, **kwargs)

    def butter(self, **kwargs):

        return quick_butter(self.time, self.data, **kwargs)

    def psd(self, **kwargs):

        return quick_psd(self.time, self.data, data_unit=self.units, **kwargs)

    def psd_welch(self, **kwargs):

        return quick_psd_welch(self.time, self.data, data_unit=self.units, **kwargs)

    def rolling_tidal_harmonic(self, Fin=None, **kwargs):

        return quick_rolling_tidal_harmonic(self.time, self.data, Fin=Fin, **kwargs)

    def tidal_harmonic(self, Fin=None, **kwargs):

        return quick_tidal_harmonic(self.time, self.data, Fin=Fin, **kwargs)

def quick_butter(time, data, T_cut_seconds, order=4, btype='lowpass', use_sos=False):
    """
    Very quick butterworth filter that works with python time objects. 
    
    Usage:
        filtered = quick_butter(time, data, T_cut_seconds, order=4, btype='lowpass')

    args:
        time: numpy array of time objects that can be converted by zutils.time.seconds_since and zutils.time.is_well_spaced
        data: np.array to be filtered. Can be higher dimensional but goes into signal.butter as is.
        T_cut_seconds: cutoff period in seconds
        
    kwargs: 
        btype: 'lowpass' or 'highpass'
        order: order of the butterworth

    """

    quick_validate(time, data)
    
    # Convert to seconds
    t_sec = ztime.seconds_since(time)

    dt = t_sec[1] - t_sec[0]
    f_s_Hz = 1/dt # my sampling freq
    f_N_Hz = 1/dt # Nyqvist frequency in function of my sampling frequency

    # Validation [should be fine for mopdel data]
    ztime.is_well_spaced(time, f_s_Hz)

    # Filter design
    if type(T_cut_seconds) == list:
        T_cut_seconds = np.array(T_cut_seconds)

    f_cut_Hz = 1/T_cut_seconds

    f_cut_norm = f_cut_Hz/f_N_Hz  # normalized cut_off frequency
    f_cut_norm = f_cut_Hz/(f_N_Hz/2)  # normalized cut_off frequency

    if np.any(f_cut_norm == 1) or np.any(f_cut_norm == 0):
        raise(Exception('Cutoff time of {} seconds is not acceptable for thge time spacing on this dataset.'.format(T_cut_seconds)))

    if order > 4:
        if not use_sos:
            use_sos = True
            print('Order greater than 4, using sos')

    if use_sos:
        # sos = signal.butter(order, Wn, btype=btype, analog=0, output='sos', fs=1)
        # out = signal.sosfiltfilt(sos, ytmp, axis=axis, padlen=0)
        # sos = signal.butter(order, f_cut_norm, btype=btype, analog=0, output='sos', fs=1)
        sos = signal.butter(order, f_cut_norm, btype=btype, output='sos')
        out = signal.sosfiltfilt(sos, data)
    else:
        #(b, a) = signal.butter(order, Wn, btype=btype, analog=0, output='ba')
        #return signal.filtfilt(b, a, ytmp, axis=-1)
        b, a = signal.butter(order, f_cut_norm, btype)
        out = signal.filtfilt(b, a, data, axis=0)

    return out

def quick_interp(time, data, **kwargs):
    """
    Very quick interp of timeseries data. 

    Usage:
        time_out_secs, time_out_np64, var_interp = quick_interp(time, data, **kwargs)

    Arguments:
        time
        data
    
    Key word arguments:
        dt_sec
        start
        end

    Outputs:
        time_out_secs
        time_out_np64 
        var_interp
    """
    dt_sec = kwargs.pop('dt_sec', 60)
    start = kwargs.pop('start', None)
    end = kwargs.pop('end', None)
    
    dt_file = time[1] - time[0]
    print('File dt initial: {}'.format(dt_file))
    print('dt out: {}'.format(dt_sec))
    
    ref_time = time[0]

    time_in_secs = ztime.seconds_since(time, ref_time)
    
    if start is None:
        start_sec = time_in_secs[0]
    else:
        start = np.datetime64(start)
        start_sec = ztime.seconds_since(start, ref_time)
        
    if end is None:
        end_sec = time_in_secs[-1]
    else:
        end = np.datetime64(end)
        end_sec = ztime.seconds_since(end, ref_time)
        end_sec += dt_sec

    time_out_secs = np.arange(start_sec, end_sec, dt_sec) 
    time_out_np64 = np.array([ref_time + np.timedelta64(int(i), 's') for i in time_out_secs])

    if len(data.shape) == 1:
        var_interp = np.interp(time_out_secs, time_in_secs, data) 
    else:
        var_interp = np.array([np.interp(time_out_secs, time_in_secs, var_row) for var_row in data])

    return time_out_secs, time_out_np64, var_interp

def quick_interp_tknown(time, data, interp_time):
    """
    Very quick interp of timeseries data. Here the time vector is known. 

    Usage:
        time_out_secs, time_out_np64, var_interp = quick_interp(time, data, interp_time)

    Arguments:
        time
        data
        interp_time
    
    Key word arguments:
        

    Outputs:
        time_out_secs
        var_interp
    """
    
    
    ref_time = time[0]

    time_in_secs = ztime.seconds_since(time, ref_time)
    time_out_secs = ztime.seconds_since(interp_time, ref_time)
    time_out_np64 = np.array([ref_time + np.timedelta64(int(i), 's') for i in time_out_secs])

    if len(data.shape) == 1:
        var_interp = np.interp(time_out_secs, time_in_secs, data) 
    else:
        var_interp = np.array([np.interp(time_out_secs, time_in_secs, var_row) for var_row in data])

    return time_out_secs, time_out_np64, var_interp

def tslice(time, start, end):
    """
    Get a quick time index for a slice between start and end dates. Does this by first converting to seconds_since. This is not fast, but it is easy for the operator as seconds_since handles the object conversion. 
    
    Usage:
        tind = tslice(time, start, end)

    Inputs:
        time: Time vector
        start: datetime.datatime np.datetime64 for the start time
        end: datetime.datatime np.datetime64 for the end time
    """

    time = ztime.seconds_since(time)
    start = ztime.seconds_since(start)
    end = ztime.seconds_since(end)
    
    tind = np.logical_and(time>=start, time<=end)
    
    return tind

def tnearest(time, target):
    """
    Get a quick time index for nearest point to target time. Does this by first converting to seconds_since. This is not fast, but it is easy for the operator as seconds_since handles the object conversion. 
    
    Usage:
        tind = tnearest(time, start, end)

    Inputs:
        time: Time vector
        target: datetime.datatime np.datetime64 for the target time
    """

    time = ztime.seconds_since(time)
    target = ztime.seconds_since(target)
    
    d = abs(time-target)
    tind = np.where(d==min(d))[0][0]
    
    return tind


def quick_psd(time, data, data_unit='m', time_unit='s', radians=False, plot=False, **kwargs):
    """
    Quick power spectral density.There are a few ways to code this. I've chosen to use scipysignal.welch
    but setting the overlap and nperseg such that there is no averaging.

    """

    quick_validate(time, data)

    plot_kwargs = kwargs.pop('plot_kwargs', {'color': 'k'})
    ax = kwargs.pop('ax', None)

    if not type(time) == np.ndarray:
        raise(Exception('Time must be an np array'))
        
    # time = ztime.seconds_since(time)
    # d_time = np.diff(time)
    # dd_time = np.abs(np.diff(d_time))

    # if not np.all(dd_time<1e-4):
    #     raise(Exception("Time must be evenly spaced"))
    t_sec = ztime.seconds_since(time)

    dt = t_sec[1] - t_sec[0]
    Fs = 1/dt # my sampling freq

    ztime.is_well_spaced(time, Fs)

    # Fs = 1/d_time[0]
    f_psd, y_psd = signal.welch(data, fs=Fs, window='hann', nperseg=len(time), noverlap=0, axis=0)

    if time_unit.lower() in ['hour', 'hours', 'h']:
        time_unit='h'
        sf = 3600
    elif time_unit.lower() in ['day', 'days', 'd']:
        time_unit='d'
        sf = 86400
    elif time_unit.lower() in ['second', 'seconds', 's']:
        time_unit='s'
        sf = 1
    else:
        raise(Exception('Time unit not recognised.'))


    if radians:
        sf *=2 *np.pi
        freq_unit = '{}$^{{-1}}$'.format(time_unit)
    else:
        freq_unit = 'cp' + time_unit

    spec_unit = '[{}]$^2$/{}'.format(data_unit, freq_unit)

    f_psd = f_psd*sf
    y_psd = y_psd/sf

    if plot:
        if len(data.shape)>1:
            print("Plotting not enabled for n dimensional data.")
        else:
            if ax is None:
                ax = plt.gca()
            # ax = plt.figure(figsize=(15, 4))
            ax.plot(f_psd, y_psd, **plot_kwargs)
            ax.set_ylabel(spec_unit)
            ax.set_xlabel(freq_unit)
            ax.set_xscale('log')
            ax.set_yscale('log')

    return f_psd, y_psd, freq_unit, spec_unit
    
def quick_psd_welch(time, data, nperseg=1024, data_unit='m', time_unit='s', radians=False, plot=False, **kwargs):
    """

    """

    quick_validate(time, data)

    plot_kwargs = kwargs.pop('plot_kwargs', {'color': 'k'})
    ax = kwargs.pop('ax', None)

    if not type(time) == np.ndarray:
        raise(Exception('Time must be an np array'))
        
    # time = ztime.seconds_since(time)
    # d_time = np.diff(time)
    # dd_time = np.abs(np.diff(d_time))

    # if not np.all(dd_time<1e-4):
    #     raise(Exception("Time must be evenly spaced"))
    t_sec = ztime.seconds_since(time)

    dt = t_sec[1] - t_sec[0]
    Fs = 1/dt # my sampling freq

    ztime.is_well_spaced(time, Fs)

    # Fs = 1/d_time[0]
    f_welch, y_welch = signal.welch(data, fs=Fs, window='hann', nperseg=nperseg, noverlap=nperseg/2, axis=0)

    if time_unit.lower() in ['hour', 'hours', 'h']:
        time_unit='h'
        sf = 3600
    elif time_unit.lower() in ['day', 'days', 'd']:
        time_unit='d'
        sf = 86400
    elif time_unit.lower() in ['second', 'seconds', 's']:
        time_unit='s'
        sf = 1
    else:
        raise(Exception('Time unit not recognised.'))


    if radians:
        sf *=2 *np.pi
        freq_unit = '{}$^{{-1}}$'.format(time_unit)
    else:
        freq_unit = 'cp' + time_unit

    spec_unit = '[{}]$^2$/{}'.format(data_unit, freq_unit)

    f_welch = f_welch*sf
    y_welch = y_welch/sf

    if plot:
        if len(data.shape)>1:
            print("Plotting not enabled for n dimensional data.")
        else:
            if ax is None:
                ax = plt.gca()
            # ax = plt.figure(figsize=(15, 4))
            ax.plot(f_welch, y_welch, **plot_kwargs)
            ax.set_ylabel(spec_unit)
            ax.set_xlabel(freq_unit)
            ax.set_xscale('log')
            ax.set_yscale('log')

    return f_welch, y_welch, freq_unit, spec_unit

def quick_psd_welch_taylor_fth(time, data, U, nperseg=1024, data_unit='m/s', radians=False, plot=False, verbose=True, **kwargs):
    """
    A quick pwelch with Taylor's frozen turbulence hypothesis. 
    
    This is simply quick_psd but takes the mean velocity U in m/s. 
    
    time_unit is no longer an option. 
    
    """

    quick_validate(time, data, verbose=verbose)

    plot_kwargs = kwargs.pop('plot_kwargs', {'color': 'k'})
    ax = kwargs.pop('ax', None)

    time_unit='s'

    if not type(time) == np.ndarray:
        raise(Exception('Time must be an np array'))
        
    time = ztime.seconds_since(time)
    d_time = np.diff(time)
    dd_time = np.abs(np.diff(d_time))

    if not np.all(dd_time<1e-4):
        raise(Exception("Time must be evenly spaced"))

    Fs = 1/d_time[0]
    f_welch, y_welch = signal.welch(data, fs=Fs, window='hann', nperseg=nperseg, noverlap=nperseg/2, axis=0)

    if time_unit.lower() in ['second', 'seconds', 's']:
        time_unit='s'
        sf = 1
    else:
        raise(Exception('Time unit not recognised.'))

    kPSD = U*y_welch
    k    = f_welch/U
    wavenumber_unit = 'm'

    if radians:
        sf *=2 *np.pi
        k_unit = '{}$^{{-1}}$'.format(wavenumber_unit)
    else:
        k_unit = 'cp' + wavenumber_unit

    kPSD_unit = '[{}]$^2$/{}'.format(data_unit, k_unit)

    k = k*sf
    kPSD = kPSD/sf
    
    out = {}
    out['k']    = k
    out['kPSD'] = kPSD
    
    out['k_unit']    = k_unit
    out['kPSD_unit'] = kPSD_unit

    if plot:
        if len(data.shape)>1:
            print("Plotting not enabled for n dimensional data.")
        else:
            if ax is None:
                ax = plt.gca()
            ax.plot(out['k'], out['kPSD'] , **plot_kwargs)
            ax.set_xlabel(k_unit)
            ax.set_ylabel(kPSD_unit)
            ax.set_xscale('log')
            ax.set_yscale('log')

    return out 

def quick_despike_gn02(time, data, intensity=[0.7], **kwargs):
    """
    Quick despike by GN02
    
    If intensity is iterable GN02 is iterated with variable intensity on each iteration. 
    
    If intensity is a number only one iteration is performed.
    """
    
    quick_validate(time, data)
    
    if not type(time) == np.ndarray:
        raise(Exception('Time must be an np array'))
        
    t_sec = ztime.seconds_since(time)
    d_time = np.diff(t_sec)
    dd_time = np.abs(np.diff(d_time))

    fs = 1/d_time[0]
    
    # if not np.all(dd_time<1e-6):
    #     raise(Exception("Time must be evenly spaced"))
    ztime.is_well_spaced(time, fs)

    fs = None      # Not actually used

    data_c = data.copy()
    any_clean = np.isnan(data_c)

    if not type(intensity) is list:
        intensity = [intensity]
    
    n_iter = len(intensity)
    for i, int_val in enumerate(intensity):
        # print('GN iteration {}'.format(i))
        # print(np.isnan(data_c))
        data_c, di = zclean.despikei(data_c, fs=fs, intensity=int_val)
        any_clean = np.logical_or(any_clean, di)
        

    return data_c, any_clean

def quick_decompose(time, data, T_cut_low, T_cut_high, plot=False, **kwargs):
    
    if 'btype' in kwargs:
        raise Exception("Can't pass btype to decompose")

    if T_cut_low >= T_cut_high:
        raise Exception("T_cut_low >= T_cut_high")

    
    data_hp = quick_butter(time, data, T_cut_low,  btype='highpass')
    data_lp = quick_butter(time, data, T_cut_high,  btype='lowpass')
    data_bp = quick_butter(time, data, [T_cut_high, T_cut_low],  btype='bandpass')

    if plot:

        quick_psd_welch(time, data,    nperseg=8*1024, plot=True, time_unit='d', plot_kwargs={'color': '0.5'})
        quick_psd_welch(time, data_lp, nperseg=8*1024, plot=True, time_unit='d', plot_kwargs={'color': 'c'})
        quick_psd_welch(time, data_bp, nperseg=8*1024, plot=True, time_unit='d', plot_kwargs={'color': 'k'})
        quick_psd_welch(time, data_hp, nperseg=8*1024, plot=True, time_unit='d', plot_kwargs={'color': 'b'})

        ylim = plt.gca().get_ylim()
        plt.plot([24*3600/(T_cut_low)]*2, ylim, 'r--')
        plt.plot([24*3600/(T_cut_high)]*2, ylim, 'r--')
        plt.gca().set_ylim(ylim)

    return data_lp, data_bp, data_hp


def quick_tidal_harmonic(time, data, Fin=None, **kwargs):

    freq, fout = sha.getTideFreq(Fin=Fin)
    # amp, phs, c0, data_fit, data_residual = 
    output = quick_harmonic(time, data, freq, **kwargs)
    output['fout'] = fout

    return output

def quick_rolling_tidal_harmonic(time, data, Fin=None, **kwargs):

    if Fin is None:
        Fin = ['M2', 'M4', 'M6']

    freq, fout = sha.getTideFreq(Fin=Fin)
    output = quick_rolling_harmonic(time, data, freq, **kwargs)
    output['fout'] = fout
    output['freq'] = freq
    # output['fout'] = fout

    return output

def quick_rolling_harmonic(time, data, omega, windowlength=3*86400.0, overlap=12*3600.0, zero_pad_nans=True, **kwargs):
    """

    kwargs:
        remove_trans_in_recon - if true this sets the window swapover point in the reconstructed harmonic signal to np.nan to limit discontinuity.

    """

    if (spec := importlib.util.find_spec('sfoda')) is not None:
        module = importlib.util.module_from_spec(spec)
        sys.modules['sfoda'] = module
        spec.loader.exec_module(module)
    else:
        raise(Exception(f"This function requires sfoda"))

    from sfoda.utils import harmonic_analysis as sha

    remove_transition_in_recon = kwargs.pop("remove_transition_in_recon", True)

    pt1, pt2 = window_index_time(time, windowlength, overlap)
    npt = len(pt1)
    nfrq = len(omega)
    amp = np.zeros((nfrq, npt,))
    phs = np.zeros((nfrq, npt,))
    c0 = np.zeros((nfrq, npt,))
    ymean = np.zeros((npt,))

    # I know there is a cleaner way than this, revisit
    if data.ndim>1:
        nother = [1, 1]+[data.shape[i] for i in np.arange(1, data.ndim)] 
        for i in np.arange(data.ndim-1):
            amp = amp[..., np.newaxis]
            phs = phs[..., np.newaxis]
            c0 = c0[..., np.newaxis]
        
        amp = amp*np.ones(nother)
        phs = phs*np.ones(nother)
        c0  = c0*np.ones(nother)

    tmid = []
    ii=-1

    data_fit = np.zeros_like(data)
    data_residual = np.zeros_like(data)

    for t1, t2 in zip(pt1,pt2):
        ii+=1

        time_ = time[t1:t2]
        data_ = data[t1:t2, ...].copy()

        # print(time_.shape)
        # print(data_.shape)

        if zero_pad_nans:
            data_[np.isnan(data_)] = 0

        # print(time_.shape)
        # print(data_.shape)

        a, p, c = sha.harmonic_fit(time_, data_, frq=omega)
        dataf_ = sha.harmonic_signal(time_, a, p, c, omega)

        if remove_transition_in_recon:
            dataf_[0] = np.nan
            dataf_[-1] = np.nan

        data_fit[t1:t2] = dataf_
        data_residual[t1:t2] = data[t1:t2] - dataf_

        amp[:, ii, ...], phs[:, ii, ...], c0[:, ii, ...] = a, p, c

        # Return the mid time point
        ind = int(np.floor(t1 + (t2-t1)/2))
        tmid.append(time[ind])
#           
        # Return the fitted time series
        ymean[ii] = data[t1:t2].mean()

    tmid = np.asarray(tmid)

    output={}
    output['tmid']              = tmid
    output['amp']               = amp
    output['phs']               = phs
    output['c0']                = c0
    output['data_fit']          = data_fit
    output['data_residual']     = data_residual
    output['omega']             = omega

    return output

def quick_harmonic(time, data, omega, zero_pad_nans=True, **kwargs):
    """

    kwargs:

    """

    if (spec := importlib.util.find_spec('sfoda')) is not None:
        module = importlib.util.module_from_spec(spec)
        sys.modules['sfoda'] = module
        spec.loader.exec_module(module)
    else:
        raise(Exception(f"This function requires sfoda"))

    from sfoda.utils import harmonic_analysis as sha

    nfrq     = len(omega)
    amp      = np.zeros((nfrq, ))
    phs      = np.zeros((nfrq, ))
    c0       = np.zeros((nfrq, ))

    # I know there is a cleaner way than this, revisit
    if data.ndim>1:
        nother = [1]+[data.shape[i] for i in np.arange(1, data.ndim)] 
        for i in np.arange(data.ndim-1):
            amp = amp[..., np.newaxis]
            phs = phs[..., np.newaxis]
            c0 = c0[..., np.newaxis]
        
        amp = amp*np.ones(nother)
        phs = phs*np.ones(nother)
        c0  = c0*np.ones(nother)

    data_fit      = np.zeros_like(data)
    data_residual = np.zeros_like(data)

    time_ = time
    data_ = data.copy()

    if zero_pad_nans:
        data_[np.isnan(data_)] = 0

    a, p, c = sha.harmonic_fit(time_, data_, frq=omega)
    dataf_  = sha.harmonic_signal(time_, a, p, c, omega)

    data_fit = dataf_
    data_residual = data - dataf_

    amp[:, ...], phs[:, ...], c0[:, ...] = a, p, c

    output={}
    output['amp']               = amp
    output['phs']               = phs
    output['c0']                = c0
    output['data_fit']          = data_fit
    output['data_residual']     = data_residual
    output['omega']             = omega

    return output
    
def quick_validate(time, data, verbose=False):

    assert type(time) in [np.array, np.ndarray], "Time must be an np.array"
    assert type(data) in [np.array, np.ndarray], "Data must be an np.array"
    assert len(time)==data.shape[0], "Time must be the same length as the first dimension of Data"

    if verbose:
        print('Validation checks passed')

def window_index_time(t, windowsize, overlap):
    """
    Determines the indices for window start and end points of a time vector
    
    The window does not need to be evenly spaced
    
    Inputs:

        t: list or array of datetime objects

        windowsize: length of the window [seconds]

        overlap: number of overlap points [seconds]
        
    Returns: pt1,pt2 the start and end indices of each window
    """
    
    tsec = ztime.seconds_since(t)
        
    t1=tsec[0]
    t2=t1 + windowsize
    pt1=[0]
    pt2=[np.searchsorted(tsec,t2)]
    while t2 < tsec[-1]:
        t1 = t2 - overlap
        t2 = t1 + windowsize

        pt1.append(np.searchsorted(tsec,t1))
        pt2.append(np.searchsorted(tsec,t2))
        
    return pt1, pt2



import sys
import os
import fileinput
import numpy as np 

from netCDF4 import Dataset
import glob

def append_and_print(dump_list, str, prnt):
    """
    Utility function that prints strings and adds them to a list. 
    """
    dump_list += [str]
    if prnt:
        print(str)
    
    return dump_list

def check_first_4_bytes(fn):
    """
    Checks first 4 bytes of a file. This can be used to identify the file type. 
    Files with user data before the header may affect this.

    On distinguishing files:

        ```
        "The short answer is that under most circumstances, you should not care, if you use version 4.0 or later of the netCDF library to access data in the file. But the difference is indicated in the first four bytes of the file, which are 'C', 'D', 'F', '\001' for the classic netCDF CDF-1 format; 'C', 'D', 'F', '\002' for the 64-bit offset CDF-2 format; 'C', 'D', 'F', '\005' for the 64-bit data CDF-5 format; or '\211', 'H', 'D', 'F' for an HDF5 file, which could be either a netCDF-4 file or a netCDF-4 classic model file. (HDF5 files may also begin with a user-block of 512, 1024, 2048, ... bytes before what is actually an 8-byte signature beginning with the 4 bytes above.)

        With netCDF version 4.0 or later, there is an easy way that will distinguish between netCDF-4 and netCDF-4 classic model files, using the "-k" option to ncdump to determine the kind of file, for example:"
        ```


    from anoher site:

        ```
        "We refer "CDF-1" as the identification string, "magic", occupying the first 4 bytes of a netCDF file. The string can be "CDF1", "CDF2", or "CDF5".

        CDF-1 and CDF-2 are also referred by the ESDS Community Standard as NetCDF Classic and 64-bit Offset File Formats, respectively. See [ESDS-RFC-011v2.0]

        The difference between CDF-1 and CDF-2 is only in the VERSION byte (\x01 vs. \x02) and the OFFSET entity, a 64-bit instead of a 32-bit offset from the beginning of the file. See CDF-2 file format specification for the detailed specifications of both CDF-1 and CDF-2.

        Below is an older version of CDF file format specification"
        ```

    """
    
    print('Tyr first 4 bytes of "' + fn + '" are:')
    with open(fn, "rb") as file:

        bytel = [file.read(1) for i in np.arange(0, 4)]
        print(bytel)
        
        if bytel == [b'C', b'D', b'F', b'\x01']:
            print('This looks like a CDF-1 file [NETCDF3_CLASSIC]')
            format = 'NETCDF3_CLASSIC'
        elif bytel == [b'C', b'D', b'F', b'\x02']:
            print('This looks like a CDF-2 file [NETCDF3_64BIT]')
            format = 'NETCDF3_64BIT'
        elif bytel[1::] ==  [b'H', b'D', b'F']:
            print('This looks like a NetCDF4 file of some type, guessing NETCDF4 but could be NETCDF4_CLASSIC')
            format = 'NETCDF4'
        else:
            print('This byte structure is unknown, might be a file type I''m not aware of. It might not be a netcdf file, or it might have a user data string.')
            format = None   
    
    print('    ')
    return format

    
def nc_compare(lhs_file, rhs_file):
    """
    Compare 2 netcdfs. Note, when readint the file contents, both are opened as 'NETCDF4' format. 
    Not sure this is the best idea.
    """
    
    print('\\ COMPARING FILETYPES:')
    format_lhs = check_first_4_bytes(lhs_file)
    format_rhs = check_first_4_bytes(rhs_file)
    
    print('\n\n\\ COMPARING REMAINING FILE DIFFERENCES (only differences are shown):')

    lhs_dump_list = ncdump(lhs_file, verbose=False, format=format_lhs)
    rhs_dump_list = ncdump(rhs_file, verbose=False, format=format_rhs)
    
    ldifferences = 0
    rdifferences = 0
    for entry in lhs_dump_list:
        if entry in rhs_dump_list:
            pass
        else:
            ldifferences += 1
            print('"{}" \n\t\t is not in the right hand file [{}]!\n'.format(entry, rhs_file))

    print(' ')

    for entry in rhs_dump_list:
        if entry in lhs_dump_list:
            pass
        else:
            rdifferences += 1
            print('"{}" \n\t\t is not in the left hand file [{}]!\n'.format(entry, lhs_file))

    print('\\ SUMMARY:')
    if format_lhs == format_rhs:
        print('  \\ File types are the same!:')
    else:
        print('  \\ WARNING: FILE TYPES NOT THE SAME!:')
        
    print('  \\ There were {} items in the RHS file not seen in the LHS file (more details above)!:'.format(rdifferences))
    print('  \\ There were {} items in the LHS file not seen in the RHS file (more details above)!:'.format(rdifferences))
    
def ncdump(filename, verbose=True, format='NETCDF4'):
    """
    Do an ncdump to screen and return all a list of strings, where each string is a line in the dump. Also checks first 4 bytes for file type. 

    Inputs:
        - filename: name of the netcdf file
        - verb: Bool controlling whether to print to screen. If false just return the list. 
        - format: format of the netcdf file: THE FUNCTION SHOULD DETERMINE THIS.

    Outputs:
        - dump_list: a list of strings, where each string is one line in the dump.  
    """

    check_first_4_bytes(filename)

    DS = Dataset(filename, mode='r', format=format)

    dump_list = ncdump_from_DS(DS, verbose=verbose)
    
    return dump_list

def ncdump_from_DS(DS, verbose=True):
    """
    Do an ncdump to screen and return all a list of strings, where each string is a line in the dump. 
    """
    
    dump_list = []
    
    dump_list = append_and_print(dump_list, 'dimensions:', verbose)

    for dimension_name in DS.dimensions.keys():
        dim = DS.dimensions[dimension_name]
        
        if dim.isunlimited():
            dump_list = append_and_print(dump_list, '\t{} = {} [UNLIMITED]'.format(dimension_name, dim.size), verbose)
        else:
            dump_list = append_and_print(dump_list, '\t{} = {}'.format(dimension_name, dim.size), verbose)

    dump_list = append_and_print(dump_list, 'variables:', verbose)

    for variable_name in DS.variables.keys():
    #     print('        {}     :   {}'.format(attr, DS.getncattr(attr)))
        var = DS.variables[variable_name]
        dimstr = ''
        for dimension in var.dimensions:
            dimstr += dimension+', '

        dimstr = dimstr[0:-2]

        dump_list = append_and_print(dump_list, '\t{} {}({})'.format(var.dtype, variable_name, dimstr), verbose)

        for attr in var.ncattrs():
            if type(var.getncattr(attr)) == str:
                dump_list = append_and_print(dump_list, '\t\t{}:{} = "{}" ;'.format(variable_name, attr, var.getncattr(attr)), verbose)
            else:
                dump_list = append_and_print(dump_list, '\t\t{}:{} = {} ;'.format(variable_name, attr, var.getncattr(attr)), verbose)


    dump_list = append_and_print(dump_list,'// global attributes:', verbose)

    for attr in DS.ncattrs():
    #     print('        {}     :   {}'.format(attr, ifile.getncattr(attr)))
        if type(DS.getncattr(attr)) == str:
            dump_list = append_and_print(dump_list, '\t\t:{} = "{}" ;'.format(attr, DS.getncattr(attr)), verbose)
        else:
            dump_list = append_and_print(dump_list, '\t\t:{} = {} ;'.format(attr, DS.getncattr(attr)), verbose)
    
    return dump_list 

def ncdump_old(filename, verbose=True, format='NETCDF4'):
    
    nc_fid = Dataset(infile, mode='r', format=format)

    ncdump_from_Dataset_old(nc_fid, verbose=True)

def ncdump_from_Dataset_old(nc_fid, verb=True):
    '''
    ncdump outputs dimensions, variables and their attribute information.
    The information is similar to that of NCAR's ncdump utility.
    ncdump requires a valid instance of Dataset.

    Parameters
    ----------
    nc_fid : netCDF4.Dataset
        A netCDF4 dateset object
    verb : Boolean
        whether or not nc_attrs, nc_dims, and nc_vars are printed

    Returns
    -------
    nc_attrs : list
        A Python list of the NetCDF file global attributes
    nc_dims : list
        A Python list of the NetCDF file dimensions
    nc_vars : list
        A Python list of the NetCDF file variables
    '''
    def print_ncattr(key):
        """
        Prints the NetCDF file attributes for a given key

        Parameters
        ----------
        key : unicode
            a valid netCDF4.Dataset.variables key
        """
        try:
            print("\t\ttype:", repr(nc_fid.variables[key].dtype))
            for ncattr in nc_fid.variables[key].ncattrs():
                print('\t\t%s:' % ncattr,\
                      repr(nc_fid.variables[key].getncattr(ncattr)))
        except KeyError:
            print("\t\tWARNING: %s does not contain variable attributes" % key)

    # NetCDF global attributes
    nc_attrs = nc_fid.ncattrs()
    if verb:
        print ("NetCDF Global Attributes:")
        for nc_attr in nc_attrs:
            print('\t%s:' % nc_attr, repr(nc_fid.getncattr(nc_attr)))
    nc_dims = [dim for dim in nc_fid.dimensions]  # list of nc dimensions
    # Dimension shape information.
    if verb:
        print("NetCDF dimension information:")
        for dim in nc_dims:
            print("\tName:", dim)
            print("\t\tsize:", len(nc_fid.dimensions[dim]))
            print_ncattr(dim)
    # Variable information.
    nc_vars = [var for var in nc_fid.variables]  # list of nc variables
    if verb:
        print("NetCDF variable information:")
        for var in nc_vars:
            if var not in nc_dims:
                print('\tName:', var)
                print("\t\tdimensions:", nc_fid.variables[var].dimensions)
                print("\t\tsize:", nc_fid.variables[var].size)
                print_ncattr(var)
    return nc_attrs, nc_dims, nc_vars



def nccopy(DS, copy_name):
    """
    The idea here is to be able to directly copy a netcdf file. It's under way - not fully implemented. 
    """

    raise(Exception('Function not ready yet.'))

    with Dataset(copy_name,'w',format='NETCDF4') as new_ncfile :

        dump_list = []

        dump_list = append_and_print(dump_list, 'dimensions:')

        for dimension_name in DS.dimensions.keys():
            dim = DS.dimensions[dimension_name]

            if dim.isunlimited():
                dump_list = append_and_print(dump_list, '\t{} = {} [UNLIMITED]'.format(dimension_name, dim.size))
            else:
                dump_list = append_and_print(dump_list, '\t{} = {}'.format(dimension_name, dim.size))

        dump_list = append_and_print(dump_list, 'variables:')

        for variable_name in DS.variables.keys():
        #     print('        {}     :   {}'.format(attr, DS.getncattr(attr)))
            var = DS.variables[variable_name]
            dimstr = ''
            for dimension in var.dimensions:
                dimstr += dimension+', '

            dimstr = dimstr[0:-2]

            dump_list = append_and_print(dump_list, '\t{} {}({})'.format(var.dtype, variable_name, dimstr))
            new_ncfile.createVariable(variable_name, var.dtype)
            
            for attr in var.ncattrs():
                if type(var.getncattr(attr)) == str:
                    dump_list = append_and_print(dump_list, '\t\t{}:{} = "{}" ;'.format(variable_name, attr, var.getncattr(attr)))
                else:
                    dump_list = append_and_print(dump_list, '\t\t{}:{} = {} ;'.format(variable_name, attr, var.getncattr(attr)))


        print('// global attributes:')

        for attr in DS.ncattrs():
        #     print('        {}     :   {}'.format(attr, ifile.getncattr(attr)))
            if type(DS.getncattr(attr)) == str:
                dump_list = append_and_print(dump_list, '\t\t:{} = "{}" ;'.format(attr, DS.getncattr(attr)))
            else:
                dump_list = append_and_print(dump_list, '\t\t:{} = {} ;'.format(attr, DS.getncattr(attr)))

        return dump_list 